/**
  Raytracing demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-01-22 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"
#include "osl/mat4.h" // basic 4x4 matrix
#include "osl/mat4_inverse.h" // 4x4 matrix inverses

void display(void) 
{
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glEnable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		1000.0); // z clipping planes
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	oglCameraAxes();
	
	/* Grid on the ground */
	double res=10.0;
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(0.1);
	glBegin(GL_LINES);
	glColor3f(0,0,0);
	for (float t=-res;t<=res;t+=1.0) {
		float z=0.0;
		glVertex3f(t,-res,z); // -Y
		glVertex3f(t,+res,z); // +Y
		glVertex3f(-res,t,z); // -X
		glVertex3f(+res,t,z); // +X
		
	}
	glEnd();
	
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.01; /* revolutions per second */
	float angle=glutGet(GLUT_ELAPSED_TIME)*0.001*rotate_rate*360;
	
	/* Build a pixel shader */
	static GLhandleARB prog=MAKE_PROGRAM_OBJECT(
		// GLSL vertex shader
		varying vec3 G; // proxy geometry location, world coordinates
		void main(void) {
			vec4 mv=gl_ModelViewMatrix*gl_Vertex; // apply modelview to get to world coords
			G=vec3(mv);
			gl_Position = gl_ProjectionMatrix*mv;
		}
	,
		// GLSL fragment shader
		uniform vec3 C; // camera position, world coordinates
		varying vec3 G; // proxy geometry location, world coordinates
		uniform vec3 reflectance; // RGB reflectance (surface color)
		
		const float miss_t = 10000.0; // t value for a miss (huge)
		const float behind_t = 0.01; // t value to consider too close to camera (avoid shadow acne)
		
		uniform mat4 A; // quadric surface matrix: 0 = x^T A x
		uniform mat4 Gmat; // converts world-coords hit to a gradient
		uniform vec3 trim_origin; // center of trimming
		// Return miss_t if this world-coordinates intersection should be trimmed.
		float trim_hit(float t,vec3 P) {
			float trim=length(P-trim_origin); // trim radius
			if (trim>1.0 || fract(6.0*trim)>0.7) {
				t=miss_t; /* cut the surface */
			}
			return t;
		}
		
		// Ray-object intersection for Quadric surface.
		//   Ray is C+t*D, return value is t, or miss_t if miss.
		//   C and D are world coordinates.
		float quadric_intersection(vec3 C,vec3 D) {
			// tack on w component
			vec4 C4=vec4(C,1.0); // w==1.0 for a position vector
			vec4 D4=vec4(D,0.0); // w==0.0 for a direction vector
			
			// matrix-vector multiply
			vec4 AD=A*D4; 
			vec4 AC=A*C4; 
			
			// Quadratic equation terms:
			float a=dot(D4,AD); // t^2 term
			float b=dot(C4,AD) + dot(D4,AC);  // t term
			float c=dot(C4,AC); // constant term
			
			if (abs(a)<0.00001) { /* NOT actually quadratic--linear only */
				float t= -c/b;  /* 0 = c + t * b, so t = -c/b */
				if (t<=behind_t) return miss_t;
				else return trim_hit(t,C+t*D);
			}

			// Solve the quadratic equation:
			float det=b*b-4.0*a*c;
			if (det<0.0) return miss_t; /* miss */
			
			// Check both roots, keep the closest root
			float closest_t=miss_t; // assume far away.
			for (float sign=-1.0;sign<=1.0;sign+=2.0)
			{
				float t=(-b+sign*sqrt(det))/(2.0*a);
				if (t>behind_t) { // valid hit, in front of camera
					t=trim_hit(t,C+t*D);
					closest_t=min(t,closest_t);
				}
			}
			return closest_t;
		}
		
		void main(void) {
			float alpha=1.0; // opacity of the final pixel
			
			// Camera ray
			vec3 D=normalize(G-C); // points from camera to proxy
			float t=quadric_intersection(C,D);
			vec3 P=C+t*D;
			
			if (t>=miss_t) alpha=0.1; // weakly show proxy geometry
			
			// Compute surface normal
			vec4 G4=Gmat*vec4(P,1.0); // convert to surface normal
			vec3 gradient=vec3(G4); // drop W coordinate
			vec3 N=normalize(gradient);
			if (dot(N,D)>0.0)
				N=-N; /* was facing wrong way... */
			
			// Compute lighting
			vec3 L=normalize(vec3(0.4,0,1.3));
			float light=clamp(dot(N,L),0.0,1.0); // diffuse
			vec3 H=normalize(0.5*(L-D)); // Blinn's halfway vector
			float spec=step(0.999,dot(H,N)); // "cartoon highlight"
			spec=pow(spec,200.0); // classic phong exponent
			
			// Shadow ray
			if (quadric_intersection(P,L)<miss_t) 
			{  /* we're in shadow */
				light*=0.2; /* much less diffuse */
				spec*=0.0; /* no specular at all */
			}
			
			light = light*0.6 + 0.4; /* ambient */
			
			vec3 rgb=light*reflectance+spec;
			gl_FragColor = vec4(rgb,alpha);
			
			// Write depth buffer value:
			vec4 P4=vec4(P,1.0); // world coords
			vec4 proj=gl_ProjectionMatrix*P4; // screen
			float depth=proj.z/proj.w; // persp. divide
			gl_FragDepth = 0.5+0.5*depth; // glDepthRange
		}
	); // end GLSL pixel shader; back to C++
	glUseProgramObjectARB(prog);
	
	glFastUniform3fv(prog,"C",1,camera);
	
	// Make a whole row of objects:
	for (float x=-1.0;x<=1.0;x+=0.25)
	for (float y=-1.0;y<=1.0;y+=0.25)
	{
		glPushMatrix();
		glTranslatef(x*10.0,y*10.0,1.0); // move to (x,y), and up Z axis a bit
		glRotatef(angle*3.0,1.0f,0.0f,1.0f); // slowly rotating

		// Quadric surface's matrix:
		osl::mat4 A(0.0f);
		A[0][0]=x; // x^2 term
		A[1][1]=y; // y^2 term
		A[2][2]=0.0; // z^2 term
		A[2][3]=1.0; // z term
		A[3][3]=0.0; // constant term

		// Move quadric by modelview matrix:
		osl::mat4 B(0.0);
		glGetFloatv(GL_MODELVIEW_MATRIX,&B[0][0]); /* read the current modelview */
		glFastUniform3fv(prog,"trim_origin",1,&B[3][0]); // object center, for trimming
		B=inverse(B); /* modelview gives obj->world xform; we need world->obj */
		osl::mat4 Amodel = transpose(B)*A*B; // rotate quadric A by matrix B
		glFastUniformMatrix4fv(prog,"A",1,0,&Amodel[0][0]);

		// Make gradient matrix:
		osl::mat4 Gmat(0.0); /* convert XYZ location to gradient */
		for (int i=0;i<4;i++) // input XYZ axis
		for (int j=0;j<3;j++) // output gradient axis
		{
			if (i==j) { // diagonal term
				Gmat[i][j] = 2.0*Amodel[i][j];
			} else {
				Gmat[i][j] = Amodel[i][j] + Amodel[j][i];
			}
		}
		glFastUniformMatrix4fv(prog,"Gmat",1,0,&Gmat[0][0]);

		glFastUniform3fv(prog,"reflectance",1,vec3(x*0.4+0.6,0.2,y*0.4+0.6)); // purple
		glutSolidSphere(1.8,8,6); // proxy geometry
		
		glPopMatrix(); // restore coordinate system
	}
	glUseProgramObjectARB(0);
	
	/*
	glColor3f(1.0,0.7,0.3); // orange wireframe sphere
	glLineWidth(2.0);
	glutWireSphere(1.3,20,30);
	*/
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("Raytrace Quadric Surface");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0,-3,2);
	
	glutMainLoop();
	return 0;
}
